from PyQt6.QtWidgets import QApplication, QMainWindow, QFileDialog


def getMainWindow() -> QMainWindow | None:
    # Global function to find the (open) QMainWindow in application
    app = QApplication.instance()
    for widget in app.topLevelWidgets():
        if isinstance(widget, QMainWindow):
            return widget
    return None


def openFileDialog(*,
    title: str | None = None,
    initial: str | None = None,
    filter: str | None = None,
    multiple: bool = False
) -> str | list[str] | None:
    parent = getMainWindow()
    if not parent: return

    if multiple:
        return QFileDialog.getOpenFileNames(
            parent=parent,
            caption=title,
            directory=initial,
            filter=filter
        )[0]

    return QFileDialog.getOpenFileName(
        parent=parent,
        caption=title,
        directory=initial,
        filter=filter
    )[0]


def openDirDialog(*,
    title: str | None = None,
    initial: str | None = None,
):
    parent = getMainWindow()
    if not parent: return

    return QFileDialog.getExistingDirectory(
        parent=parent,
        caption=title,
        directory=initial,
    )



from PyQt6.QtCore import Qt
from PyQt6.QtGui import QDoubleValidator, QIntValidator, QValidator


class Align:
    Default = Left = Qt.AlignmentFlag.AlignLeft
    Leading = Qt.AlignmentFlag.AlignLeading
    Right = Qt.AlignmentFlag.AlignRight
    Trailing = Qt.AlignmentFlag.AlignTrailing
    HCenter = Qt.AlignmentFlag.AlignHCenter
    Justify = Qt.AlignmentFlag.AlignJustify
    Absolute = Qt.AlignmentFlag.AlignAbsolute
    Horizontal_Mask = Qt.AlignmentFlag.AlignHorizontal_Mask
    Top = Qt.AlignmentFlag.AlignTop
    Bottom = Qt.AlignmentFlag.AlignBottom
    VCenter = Qt.AlignmentFlag.AlignVCenter
    Vertical_Mask = Qt.AlignmentFlag.AlignVertical_Mask
    Center = Qt.AlignmentFlag.AlignCenter
    Baseline = Qt.AlignmentFlag.AlignBaseline


def numericLimitRule(*, double: bool = False, start: float = 0, end: float = None) -> QValidator:
    val = QDoubleValidator() if double else QIntValidator()
    if start is not None: val.setBottom(start)
    if end is not None: val.setTop(end)
    return val



from typing import Unpack
from PyQt6.QtWidgets import QLabel
from PyQt6.QtGui import QMovie, QImage, QPixmap

from .label import label, LabelKwargs, Align


def gifView(
    path: str,
    **extra: Unpack[LabelKwargs]
) -> QLabel:
    container: QLabel = label(
        **extra,
        scaledContents=True,
    )
    
    movie = QMovie(path)
    container.setMovie(movie)
    movie.start()
    return container


def pictureView(
    src: str | bytes,
    **extra: Unpack[LabelKwargs]
) -> QLabel:
    container: QLabel = label(
        **extra,
        scaledContents=True,
    )

    image = QImage()
    if isinstance(src, str):
        image.load(src)
    else:
        image.loadFromData(src)

    pixmap = QPixmap.fromImage(image)
    del image
    container.setPixmap(pixmap)

    return container



from typing import NotRequired, Callable, Unpack

from PyQt6.QtWidgets import QTableWidget, QTableWidgetItem, QWidget

from .widget import WidgetKwargs, widget


class TableWidgetKwargs(WidgetKwargs):
    rowCount: NotRequired[int]
    columnCount: NotRequired[int]
    sortingEnabled: NotRequired[bool]
    showGrid: NotRequired[bool]
    cornerButtonEnabled: NotRequired[bool]

    horizontalHeaderLabels: NotRequired[list[str]]
    verticalHeaderLabels: NotRequired[list[str]]
    columnWidths: NotRequired[dict[int, int]]
    rowHeights: NotRequired[dict[int, int]]
    hiddenColumns: NotRequired[list[int]]
    hiddenRows: NotRequired[list[int]]
    rows: NotRequired[list[list[str | QWidget]]]

    onCellClicked: NotRequired[Callable[[int, int], None]]
    onCellDoubleClicked: NotRequired[Callable[[int, int], None]]
    onCellChanged: NotRequired[Callable[[int, int], None]]
    onCurrentCellChanged: NotRequired[Callable[[int, int, int, int], None]]
    onItemClicked: NotRequired[Callable[[object], None]]
    onItemDoubleClicked: NotRequired[Callable[[object], None]]
    onItemChanged: NotRequired[Callable[[object], None]]


def table(*, _table: QTableWidget | None = None, **kwargs: Unpack[TableWidgetKwargs]) -> QTableWidget:
    if _table is None:
        _table = QTableWidget()

    if (v := kwargs.get("rowCount")) is not None:
        _table.setRowCount(v)
    if (v := kwargs.get("columnCount")) is not None:
        _table.setColumnCount(v)
    if (v := kwargs.get("sortingEnabled")) is not None:
        _table.setSortingEnabled(v)
    if (v := kwargs.get("showGrid")) is not None:
        _table.setShowGrid(v)
    if (v := kwargs.get("cornerButtonEnabled")) is not None:
        _table.setCornerButtonEnabled(v)

    # Headers
    if (v := kwargs.get("horizontalHeaderLabels")) is not None:
        _table.setHorizontalHeaderLabels(v)
    if (v := kwargs.get("verticalHeaderLabels")) is not None:
        _table.setVerticalHeaderLabels(v)

    # Column widths
    if (v := kwargs.get("columnWidths")) is not None:
        for col, width in v.items():
            _table.setColumnWidth(col, width)

    # Row heights
    if (v := kwargs.get("rowHeights")) is not None:
        for row, height in v.items():
            _table.setRowHeight(row, height)

    # Hide/show rows/columns
    if (v := kwargs.get("hiddenColumns")) is not None:
        for col in v:
            _table.hideColumn(col)

    if (v := kwargs.get("hiddenRows")) is not None:
        for row in v:
            _table.hideRow(row)

    if (rows := kwargs.get("rows")) is not None:
        for r in rows:
            tableRow(_table, *r)

    # Signals
    if (fn := kwargs.get("onCellClicked")) is not None:
        _table.cellClicked.connect(fn)
    if (fn := kwargs.get("onCellDoubleClicked")) is not None:
        _table.cellDoubleClicked.connect(fn)
    if (fn := kwargs.get("onCellChanged")) is not None:
        _table.cellChanged.connect(fn)
    if (fn := kwargs.get("onCurrentCellChanged")) is not None:
        _table.currentCellChanged.connect(fn)
    if (fn := kwargs.get("onItemClicked")) is not None:
        _table.itemClicked.connect(fn)
    if (fn := kwargs.get("onItemDoubleClicked")) is not None:
        _table.itemDoubleClicked.connect(fn)
    if (fn := kwargs.get("onItemChanged")) is not None:
        _table.itemChanged.connect(fn)

    widget(**kwargs, _widget=_table)
    return _table


def tableRow(table: QTableWidget, *cells: str | QWidget, index: int | None = None):
    if index is None:
        index = table.rowCount()
    table.insertRow(index)

    for i, el in enumerate(cells):
        if isinstance(el, str):
            table.setItem(index, i, QTableWidgetItem(el))
        elif isinstance(el, QWidget):
            table.setCellWidget(index, i, el)

    return index



import sys, inspect
from pathlib import Path
from .property import classproperty


class AppStatic:
    __is_executable: bool = getattr(sys, 'frozen', False)
    __base_path: Path = None
    __assets_path: Path = None
    __data_path: Path = None
    
    @classmethod
    def init(
        cls, *,
        datapath: bool | str = False,
        assets_folder: str = "assets",
        _pick_step: int = 0,
    ):
        assert cls.__base_path is None, "Action allowed only once!"
        assert _pick_step in range(0, 5), "Not too more"
        
        if cls.is_executable:
            cls.__base_path = Path(sys.executable).absolute().parent
            cls.__assets_path = Path(sys._MEIPASS, assets_folder).absolute()
        else:
            stack = inspect.stack()
            cls.__base_path = Path(str(stack[_pick_step+1][1])).absolute().parent
            cls.__assets_path = cls.__base_path / assets_folder

        if datapath:
            cls.__data_path = cls.__base_path / (datapath if isinstance(datapath, str) else "_data")
            if not cls.__data_path.exists():
                cls.__data_path.mkdir(parents=True, exist_ok=True)

    def __init__(self) -> None:
        raise Exception("Not supported, use 'init'")
    
    @classproperty
    def is_executable(cls):
        return cls.__is_executable

    @classproperty
    def base_path(cls):
        return cls.__base_path
    
    @classproperty
    def assets_path(cls):
        return cls.__assets_path
    
    @classproperty
    def data_path(cls):
        assert cls.__data_path, "Data directory not ininitialized!"
        return cls.__data_path



from .dialogs import getMainWindow, openDirDialog, openFileDialog
from .popup_widget import errorPopup
from .overlay_feature import OverlayFeature
from .overlay_mixin import OverlayMixin, OverlayModal



from typing import TypedDict, NotRequired, Unpack

from PyQt6.QtWidgets import QWidget, QBoxLayout, QLayout

from .options import Align

BoxLayoutDirection = QBoxLayout.Direction


class BoxLayoutKwargs(TypedDict):
    spacing: NotRequired[int]
    contentsMargins: NotRequired[tuple[int, int, int, int]]  # left, top, right, bottom
    alignment: NotRequired[Align]
    direction: NotRequired[BoxLayoutDirection]
    sizeConstraint: NotRequired[QLayout.SizeConstraint]
    stretchFactors: NotRequired[dict[int, int]]  # (index, stretch)
    widgets: NotRequired[list[QWidget]]


def boxLayout(*, _layout: QBoxLayout | None = None, **kwargs: Unpack[BoxLayoutKwargs]) -> QBoxLayout:
    if _layout is None:
        _layout = QBoxLayout(kwargs.get("direction") or BoxLayoutDirection.TopToBottom)
    elif (v := kwargs.get("direction")) is not None:
        _layout.setDirection(v)

    if (v := kwargs.get("spacing")) is not None:
        _layout.setSpacing(v)

    if (v := kwargs.get("contentsMargins")) is not None:
        _layout.setContentsMargins(*v)

    if (v := kwargs.get("sizeConstraint")) is not None:
        _layout.setSizeConstraint(v)

    alignment = kwargs.get("alignment")

    if (widgets := kwargs.get("widgets")) is not None:
        for w in widgets:
            if alignment:
                _layout.addWidget(w, alignment=alignment)
            else:
                _layout.addWidget(w)

    if (v := kwargs.get("stretchFactors")) is not None:
        for index, stretch in v.items():
            _layout.setStretch(index, stretch)

    return _layout



from typing import NotRequired, TypedDict, Unpack, Callable
from PyQt6.QtCore import QRect, QSize, Qt, QObject, QPoint
from PyQt6.QtGui import QCursor, QFont, QPalette, QIcon
from PyQt6.QtWidgets import QWidget, QLayout, QSizePolicy


class WidgetKwargs(TypedDict):
    id: NotRequired[str]
    parent: NotRequired[QWidget]

    # SIZE
    fixedSize: NotRequired[QSize | tuple[int, int] | int]
    minimumSize: NotRequired[QSize | tuple[int, int]]
    maximumSize: NotRequired[QSize | tuple[int, int]]
    baseSize: NotRequired[QSize | tuple[int, int]]
    sizeIncrement: NotRequired[QSize | tuple[int, int]]
    sizePolicy: NotRequired[QSizePolicy]

    minimumWidth: NotRequired[int]
    minimumHeight: NotRequired[int]
    maximumWidth: NotRequired[int]
    maximumHeight: NotRequired[int]

    # GEOMETRY
    geometry: NotRequired[QRect | tuple[int, int, int, int]]
    pos: NotRequired[tuple[int, int]]
    x: NotRequired[int]
    y: NotRequired[int]
    width: NotRequired[int]
    height: NotRequired[int]
    size: NotRequired[QSize | tuple[int, int] | int]

    # BEHAVIOR
    enabled: NotRequired[bool]
    disabled: NotRequired[bool]
    visible: NotRequired[bool]
    hidden: NotRequired[bool]
    updatesEnabled: NotRequired[bool]
    acceptDrops: NotRequired[bool]
    mouseTracking: NotRequired[bool]
    tabletTracking: NotRequired[bool]
    focus: NotRequired[bool]

    # LOOK
    styleSheet: NotRequired[str]
    cursor: NotRequired[QCursor]
    font: NotRequired[QFont | tuple[str, int]]
    palette: NotRequired[QPalette]
    autoFillBackground: NotRequired[bool]

    # WINDOW PROPS
    windowTitle: NotRequired[str]
    windowOpacity: NotRequired[float]
    windowIcon: NotRequired[str]
    windowIconText: NotRequired[str]
    windowRole: NotRequired[str]
    windowFilePath: NotRequired[str]
    windowFlag: NotRequired[Qt.WindowType]
    windowFlags: NotRequired[Qt.WindowType]

    # LAYOUT
    layout: NotRequired[QLayout]
    contentsMargins: NotRequired[tuple[int, int, int, int]]
    layoutDirection: NotRequired[Qt.LayoutDirection]

    # INFO TEXT
    toolTip: NotRequired[str]
    statusTip: NotRequired[str]
    whatsThis: NotRequired[str]
    accessibleName: NotRequired[str]
    accessibleDescription: NotRequired[str]
    contextMenuPolicy: NotRequired[Qt.ContextMenuPolicy]

    # QWidget event handlers
    onDestroyed: NotRequired[Callable[[QObject], None]]
    onCustomContextMenuRequested: NotRequired[Callable[[QPoint], None]]


def widget(*, _widget: QWidget | None = None, **kwargs: Unpack[WidgetKwargs]) -> QWidget:
    if _widget is None:
        _widget = QWidget()

    # OBJECT NAME
    if "id" in kwargs:
        _widget.setObjectName(kwargs["id"])

    if "parent" in kwargs:
        _widget.setParent(kwargs["parent"])

    # SIZE
    if (size := kwargs.get("fixedSize")) is not None:
        if isinstance(size, tuple):
            _widget.setFixedSize(size[0], size[1])
        elif isinstance(size, int):
            _widget.setFixedSize(size, size)
        elif isinstance(size, QSize):
            _widget.setFixedSize(size)

    if "minimumSize" in kwargs:
        v = kwargs["minimumSize"]
        _widget.setMinimumSize(*v) if isinstance(v, tuple) else _widget.setMinimumSize(v)
    if "maximumSize" in kwargs:
        v = kwargs["maximumSize"]
        _widget.setMaximumSize(*v) if isinstance(v, tuple) else _widget.setMaximumSize(v)
    if "baseSize" in kwargs:
        v = kwargs["baseSize"]
        _widget.setBaseSize(*v) if isinstance(v, tuple) else _widget.setBaseSize(v)
    if "sizeIncrement" in kwargs:
        v = kwargs["sizeIncrement"]
        _widget.setSizeIncrement(*v) if isinstance(v, tuple) else _widget.setSizeIncrement(v)
    if "sizePolicy" in kwargs:
        _widget.setSizePolicy(kwargs["sizePolicy"])

    if "minimumWidth" in kwargs:
        _widget.setMinimumWidth(kwargs["minimumWidth"])
    if "minimumHeight" in kwargs:
        _widget.setMinimumHeight(kwargs["minimumHeight"])
    if "maximumWidth" in kwargs:
        _widget.setMaximumWidth(kwargs["maximumWidth"])
    if "maximumHeight" in kwargs:
        _widget.setMaximumHeight(kwargs["maximumHeight"])

    # GEOMETRY
    if "geometry" in kwargs:
        v = kwargs["geometry"]
        _widget.setGeometry(QRect(*v)) if isinstance(v, tuple) else _widget.setGeometry(v)

    if "pos" in kwargs:
        _widget.move(*kwargs["pos"])
    if "x" in kwargs:
        _widget.move(kwargs["x"], _widget.y())
    if "y" in kwargs:
        _widget.move(_widget.x(), kwargs["y"])
    if "width" in kwargs:
        _widget.resize(kwargs["width"], _widget.height())
    if "height" in kwargs:
        _widget.resize(_widget.width(), kwargs["height"])

    if (size := kwargs.get("size")) is not None:
        if isinstance(size, tuple):
            _widget.resize(size[0], size[1])
        elif isinstance(size, int):
            _widget.resize(size, size)
        elif isinstance(size, QSize):
            _widget.resize(size)

    # BEHAVIOR
    if "enabled" in kwargs:
        _widget.setEnabled(kwargs["enabled"])
    if "disabled" in kwargs:
        _widget.setDisabled(kwargs["disabled"])
    if "visible" in kwargs:
        _widget.setVisible(kwargs["visible"])
    if "hidden" in kwargs:
        _widget.setHidden(kwargs["hidden"])
    if "updatesEnabled" in kwargs:
        _widget.setUpdatesEnabled(kwargs["updatesEnabled"])
    if "acceptDrops" in kwargs:
        _widget.setAcceptDrops(kwargs["acceptDrops"])
    if "mouseTracking" in kwargs:
        _widget.setMouseTracking(kwargs["mouseTracking"])
    if "tabletTracking" in kwargs:
        _widget.setTabletTracking(kwargs["tabletTracking"])
    if "focus" in kwargs and kwargs["focus"]:
        _widget.setFocus()

    # LOOK
    if "styleSheet" in kwargs:
        _widget.setStyleSheet(kwargs["styleSheet"])
    if "cursor" in kwargs:
        _widget.setCursor(kwargs["cursor"])
    if (font := kwargs.get("font")) is not None:
        if isinstance(font, tuple):
            font = QFont(*font)
        if isinstance(font, QFont):
            _widget.setFont(font)
    if "palette" in kwargs:
        _widget.setPalette(kwargs["palette"])
    if "autoFillBackground" in kwargs:
        _widget.setAutoFillBackground(kwargs["autoFillBackground"])

    # WINDOW PROPS
    if "windowTitle" in kwargs:
        _widget.setWindowTitle(kwargs["windowTitle"])
    if "windowOpacity" in kwargs:
        _widget.setWindowOpacity(kwargs["windowOpacity"])
    if "windowIcon" in kwargs:
        _widget.setWindowIcon(QIcon(kwargs["windowIcon"]))
    if "windowIconText" in kwargs:
        _widget.setWindowIconText(kwargs["windowIconText"])
    if "windowRole" in kwargs:
        _widget.setWindowRole(kwargs["windowRole"])
    if "windowFilePath" in kwargs:
        _widget.setWindowFilePath(kwargs["windowFilePath"])
    if "windowFlag" in kwargs:
        _widget.setWindowFlag(kwargs["windowFlag"])
    if "windowFlags" in kwargs:
        _widget.setWindowFlags(kwargs["windowFlags"])

    # LAYOUT
    if "layout" in kwargs:
        _widget.setLayout(kwargs["layout"])
    if "contentsMargins" in kwargs:
        _widget.setContentsMargins(*kwargs["contentsMargins"])
    if "layoutDirection" in kwargs:
        _widget.setLayoutDirection(kwargs["layoutDirection"])

    # TEXT INFO
    if "toolTip" in kwargs:
        _widget.setToolTip(kwargs["toolTip"])
    if "statusTip" in kwargs:
        _widget.setStatusTip(kwargs["statusTip"])
    if "whatsThis" in kwargs:
        _widget.setWhatsThis(kwargs["whatsThis"])
    if "accessibleName" in kwargs:
        _widget.setAccessibleName(kwargs["accessibleName"])
    if "accessibleDescription" in kwargs:
        _widget.setAccessibleDescription(kwargs["accessibleDescription"])
    if "contextMenuPolicy" in kwargs:
        _widget.setContextMenuPolicy(kwargs["contextMenuPolicy"])

    # ---------- EVENT HANDLERS ----------
    if (fn := kwargs.get("onDestroyed")) is not None:
        _widget.destroyed.connect(fn)
    if (fn := kwargs.get("onCustomContextMenuRequested")) is not None:
        _widget.customContextMenuRequested.connect(fn)

    return _widget



from .button import button
from .divider import divider
from .image_views import gifView, pictureView
from .label import label
from .box_layout import boxLayout, BoxLayoutDirection
from .line_edit import lineEdit
from .options import Align, numericLimitRule
from .scroll import scroll
from .table import table, tableRow
from .tabs import tabs
from .widget import widget
from .window import MainWindow



from typing import Unpack, NotRequired, Callable

from PyQt6.QtGui import QValidator
from PyQt6.QtWidgets import QLineEdit

from .options import Align
from .widget import WidgetKwargs, widget


class LineEditKwargs(WidgetKwargs):
    text: NotRequired[str]
    placeholderText: NotRequired[str]
    readOnly: NotRequired[bool]
    maxLength: NotRequired[int]
    echoMode: NotRequired[QLineEdit.EchoMode]
    alignment: NotRequired[Align]
    clearButtonEnabled: NotRequired[bool]
    cursorPosition: NotRequired[int]
    inputMask: NotRequired[str]
    validator: NotRequired[QValidator]
    dragEnabled: NotRequired[bool]

    onTextChanged: NotRequired[Callable[[str], None]]
    onTextEdited: NotRequired[Callable[[str], None]]
    onEditingFinished: NotRequired[Callable[[], None]]
    onReturnPressed: NotRequired[Callable[[], None]]


def lineEdit(*, _edit: QLineEdit | None = None, **kwargs: Unpack[LineEditKwargs]) -> QLineEdit:
    if _edit is None:
        _edit = QLineEdit()

    # ---- QLineEdit parameters ----

    if (v := kwargs.get("text")) is not None:
        _edit.setText(v)

    if (v := kwargs.get("placeholderText")) is not None:
        _edit.setPlaceholderText(v)

    if (v := kwargs.get("readOnly")) is not None:
        _edit.setReadOnly(v)

    if (v := kwargs.get("maxLength")) is not None:
        _edit.setMaxLength(v)

    if (v := kwargs.get("echoMode")) is not None:
        _edit.setEchoMode(v)

    if (v := kwargs.get("alignment")) is not None:
        _edit.setAlignment(v)

    if (v := kwargs.get("clearButtonEnabled")) is not None:
        _edit.setClearButtonEnabled(v)

    if (v := kwargs.get("cursorPosition")) is not None:
        _edit.setCursorPosition(v)

    if (v := kwargs.get("inputMask")) is not None:
        _edit.setInputMask(v)

    if (v := kwargs.get("validator")) is not None:
        _edit.setValidator(v)

    if (v := kwargs.get("dragEnabled")) is not None:
        _edit.setDragEnabled(v)

    # -------- Handlers (explicit) ---------

    if (fn := kwargs.get("onTextChanged")) is not None:
        _edit.textChanged.connect(fn)

    if (fn := kwargs.get("onTextEdited")) is not None:
        _edit.textEdited.connect(fn)

    if (fn := kwargs.get("onEditingFinished")) is not None:
        _edit.editingFinished.connect(fn)

    if (fn := kwargs.get("onReturnPressed")) is not None:
        _edit.returnPressed.connect(fn)

    # ---- apply base QWidget params ----
    widget(**kwargs, _widget=_edit)

    return _edit



from PyQt6.QtGui import QResizeEvent
from PyQt6.QtWidgets import QMainWindow, QLayout, QWidget

from pyqtexa import widgets
from pyqtexa.widgets.scroll import scroll, ScrollAreaKwargs


class OverlayFeature:
    def __init__(self, window: QMainWindow) -> None:
        self.__window = window
        self.__modalsQueue: list[QWidget] = []

    @property
    def shown(self):
        return len(self.__modalsQueue) > 0

    def hide(self, *, every: bool = False):
        assert self.shown, "Modal not shown"
        if every:
            for widget in self.__modalsQueue:
                widget.setParent(None)
                widget.deleteLater()
        else:
            widget: QWidget = self.__modalsQueue.pop()
            widget.setParent(None)
            widget.deleteLater()

    def show(
        self, layout: QLayout, *,
        closeManually: bool = True,
        closeEvery: bool = False,
        scrollKwargs: ScrollAreaKwargs | None = None,
    ):
        root: QWidget = self.__window.centralWidget()

        overflow = scroll(
            **scrollKwargs,
            id="overlay-scrollarea",
            parent=root,
            widgetResizable=True,
            styleSheet="QScrollArea#overlay-scrollarea{ background: transparent; }",
            widget=widgets.widget(
                id="overflow",
                styleSheet="#overflow{ background-color:rgba(0,0,0,0." +
                           ("1" if len(self.__modalsQueue) >= 1 else "3") + "); }",
                layout=widgets.boxLayout(
                    alignment=widgets.Align.Center,
                    contentsMargins=(50, 50, 50, 50),
                    widgets=[
                        widgets.widget(
                            id="overflow-modal",
                            layout=layout,
                            styleSheet="#overflow-modal{ background-color: white; border-radius: 10px; }"
                        )
                    ]
                )
            ),
        )

        if closeManually:
            overflow.mousePressEvent = lambda ev: self.hide(every=closeEvery)

        overflow.move(0, 0)
        overflow.resize(root.size())

        self.__modalsQueue.append(overflow)

    def onResizeWindow(self, ev: QResizeEvent):
        if not self.shown: return
        for widget in self.__modalsQueue:
            widget.resize(ev.size())



from typing import Unpack, Callable, Iterable, Generic, TypeVar
from ..widgets.table import table as tableWidget, tableRow, TableWidgetKwargs
from ..widgets import button as buttonWidget
from enum import Enum
from decimal import Decimal


T = TypeVar('T')
class TableProxy(Generic[T]):
    
    STORED_ACTIONS = {
        "remove": lambda helper, obj: helper.remove(obj)
    }
    
    def __init__(
        self,
        pk: str,
        field_mapping: dict[str, str], *,
        actions: dict[str, Callable[['TableProxy', T], None] | str] = None,
        initial: Iterable[T] = None,
        handle_remove: Callable[[T], None] = None,
        **kwargs: Unpack[TableWidgetKwargs]
    ) -> None:
        self.__pk: int = pk
        self.__pk_stack = []
        self.__mapping = field_mapping
        self.__actions = actions
        self.__on_remove = handle_remove
        columns = list(field_mapping.values())
        if actions:
            for name, handler in actions.items():
                if isinstance(handler, str):
                    actions[name] = self.STORED_ACTIONS[handler]
                columns.append("")
                
        self.__widget = tableWidget(**kwargs, horizontalHeaderLabels=columns)
        if initial:
            self.extend(initial)

    @property
    def widget(self):
        return self.__widget
    
    def __validate_pk(self, obj: T):
        if not hasattr(obj, self.__pk):
            raise ValueError(f"Object '{obj.__class__.__name__}' has no pk attribute '{self.__pk}'")
        
        pk = getattr(obj, self.__pk)
        if not isinstance(pk, (int, float, str)):
            raise ValueError(f"Object '{obj.__class__.__name__}' pk attribute '{self.__pk}' bad type")
        return pk
    
    def extend(self, objs: Iterable[T]):
        for obj in objs:
            self.add(obj)
    
    def add(self, obj: T):
        pk = self.__validate_pk(obj)
        if pk in self.__pk_stack:
            return obj

        cells = []
        for name in self.__mapping.keys():
            if not hasattr(obj, name):
                raise ValueError(f"Object '{obj.__class__.__name__}' has no attribute '{name}'")
            value = getattr(obj, name)
            if isinstance(value, (int, float, Decimal, bool)):
                value = str(value)
            elif isinstance(value, Enum):
                if isinstance(value, str):
                    value = value.value
                else:
                    value = value.name
            
            cells.append(value if isinstance(value, str) else None)
        
        if self.__actions:
            for name, handler in self.__actions.items():
                cells.append(buttonWidget(text=name, onClicked=lambda _: handler(self, obj)))
        
        tableRow(self.__widget, *cells)
        self.__pk_stack.append(pk)
        return obj
    
    def remove(self, obj: T):
        pk = self.__validate_pk(obj)
        index = self.__pk_stack.index(pk)
        self.__widget.removeRow(index)
        self.__pk_stack.pop(index)
        if self.__on_remove:
            self.__on_remove(obj)
        return obj

    def clear(self):
        self.__widget.clear()
        self.__pk_stack.clear()



from PyQt6.QtWidgets import QWidget, QLabel
from PyQt6.QtCore import Qt, QPoint, QTimer

from pyqtexa.widgets import label


def errorPopup(parent: QWidget, text: str, *, ms_delay: int = 2000):
    assert parent and text and ms_delay > 100
    widget: QLabel = label(
        text=text,
        parent=parent,
        styleSheet="border: 2px solid #e31a17; border-radius: 5px;" \
        "padding: 2px; background-color: white; font-size: 14px;"
    )
    widget.setWindowFlags(Qt.WindowType.ToolTip)
    widget.move(parent.mapToGlobal(QPoint()))
    widget.show()
    QTimer.singleShot(ms_delay, widget.hide)



from typing import NotRequired, Callable, Unpack

from PyQt6.QtWidgets import QTabWidget, QWidget, QTabBar

from .widget import WidgetKwargs, widget


class TabWidgetKwargs(WidgetKwargs):
    tabsClosable: NotRequired[bool]
    movable: NotRequired[bool]
    tabPosition: NotRequired[QTabWidget.TabPosition]
    tabShape: NotRequired[QTabWidget.TabShape]
    currentIndex: NotRequired[int]

    tabBar: NotRequired[QTabBar]
    tabs: NotRequired[list[tuple[str, QWidget]]]

    # Signals
    onCurrentChanged: NotRequired[Callable[[int], None]]
    onTabCloseRequested: NotRequired[Callable[[int], None]]
    onTabBarClicked: NotRequired[Callable[[int], None]]
    onTabBarDoubleClicked: NotRequired[Callable[[int], None]]
    onTabMoved: NotRequired[Callable[[int, int], None]]


def tabs(*, _tabs: QTabWidget | None = None, **kwargs: Unpack[TabWidgetKwargs]) -> QTabWidget:
    if _tabs is None:
        _tabs = QTabWidget()

    # QTabWidget parameters
    if (v := kwargs.get("tabsClosable")) is not None:
        _tabs.setTabsClosable(v)

    if (v := kwargs.get("movable")) is not None:
        _tabs.setMovable(v)

    if (v := kwargs.get("tabPosition")) is not None:
        _tabs.setTabPosition(v)

    if (v := kwargs.get("tabShape")) is not None:
        _tabs.setTabShape(v)

    if (v := kwargs.get("currentIndex")) is not None:
        _tabs.setCurrentIndex(v)

    if (v := kwargs.get("tabBar")) is not None:
        _tabs.setTabBar(v)

    if (v := kwargs.get("tabs")) is not None:
        for label, wdg in v:
            _tabs.addTab(wdg, label)

    # Signals
    if (fn := kwargs.get("onCurrentChanged")) is not None:
        _tabs.currentChanged.connect(fn)

    if (fn := kwargs.get("onTabCloseRequested")) is not None:
        _tabs.tabCloseRequested.connect(fn)

    if (fn := kwargs.get("onTabBarClicked")) is not None:
        _tabs.tabBar().tabBarClicked.connect(fn)

    if (fn := kwargs.get("onTabBarDoubleClicked")) is not None:
        _tabs.tabBar().tabBarDoubleClicked.connect(fn)

    if (fn := kwargs.get("onTabMoved")) is not None:
        _tabs.tabBar().tabMoved.connect(fn)

    # Base QWidget parameters
    widget(**kwargs, _widget=_tabs)

    return _tabs



from abc import ABC, abstractmethod

from PyQt6.QtGui import QResizeEvent
from PyQt6.QtWidgets import QMainWindow, QLayout

from .overlay_feature import OverlayFeature
from pyqtexa import widgets
from pyqtexa.widgets.scroll import ScrollAreaKwargs


class OverlayModal(ABC):

    def __init__(self) -> None:
        assert isinstance(self, QLayout), "Mixin must be extends from QLayout"

    @abstractmethod
    def overlaySignal(self, **kwargs):
        pass

    def hide(self, *, every: bool = False):
        mixin: 'OverlayMixin' = self.window
        mixin.hideModal(every=every)


class OverlayMixin:
    def __init__(self) -> None:
        assert isinstance(self, QMainWindow), "Mixin must be extends from QMainWindow"
        self.overlayFeature = OverlayFeature(self)

    def showModal(
        self,
        content: QLayout | type[OverlayModal] | OverlayModal, *,
        closeManually: bool = True,
        closeEvery: bool = False,
        scrollKwargs: ScrollAreaKwargs | None = None,
        signal: dict = None
    ):
        if isinstance(content, type(OverlayModal)):
            content = content(self)
        if isinstance(content, OverlayModal) and signal:
            content.overlaySignal(**signal)

        self.overlayFeature.show(
            content,
            closeManually=closeManually,
            closeEvery=closeEvery,
            scrollKwargs=scrollKwargs,
        )

    def hideModal(self, *, every: bool = False):
        self.overlayFeature.hide(every=every)

    def showTextPopup(self, text: str, *, okayText: str = "OK"):
        self.showModal(
            widgets.boxLayout(
                widgets=[
                    widgets.label(text=text),
                    widgets.button(text=okayText, onClicked= lambda _: self.hideModal())
                ]
            )
        )

    @property
    def overlayShown(self):
        return self.overlayFeature.shown

    def resizeEvent(self, event: QResizeEvent):
        super(QMainWindow, self).resizeEvent(event)
        if self.overlayFeature.shown:
            self.overlayFeature.onResizeWindow(event)



from typing import Unpack

from PyQt6.QtWidgets import QFrame

from .widget import WidgetKwargs, widget


def divider(vertical: bool, **kwargs: Unpack[WidgetKwargs]) -> QFrame:
    line = QFrame()
    if vertical:
        line.setGeometry(0, 0, 3, 50)
        line.setFrameShape(QFrame.Shape.VLine)
        line.setFrameShadow(QFrame.Shadow.Sunken)
    else:
        line.setGeometry(0, 0, 50, 3)
        line.setFrameShape(QFrame.Shape.HLine)
        line.setFrameShadow(QFrame.Shadow.Sunken)

    # ---- apply base QWidget params ----
    widget(**kwargs, _widget=line)
    return line



from typing import NotRequired, Unpack

from PyQt6.QtCore import Qt
from PyQt6.QtWidgets import QScrollArea, QFrame, QWidget

from .widget import WidgetKwargs, widget


class ScrollAreaKwargs(WidgetKwargs):
    widget: NotRequired[QWidget]
    widgetResizable: NotRequired[bool]
    horizontalScrollBar: NotRequired[bool | Qt.ScrollBarPolicy]
    verticalScrollBar: NotRequired[bool | Qt.ScrollBarPolicy]
    frameShape: NotRequired[QFrame.Shape]
    frameShadow: NotRequired[QFrame.Shadow]


def scroll(*, _scroll: QScrollArea | None = None, **kwargs: Unpack[ScrollAreaKwargs]) -> QScrollArea:
    if _scroll is None:
        _scroll = QScrollArea()

    if (v := kwargs.get("widget")) is not None:
        _scroll.setWidget(v)
    if (v := kwargs.get("widgetResizable")) is not None:
        _scroll.setWidgetResizable(v)
    if (v := kwargs.get("horizontalScrollBar")) is not None:
        _scroll.setHorizontalScrollBarPolicy(
            (Qt.ScrollBarPolicy.ScrollBarAsNeeded if v else Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
            if isinstance(v, bool) else v
        )
    if (v := kwargs.get("verticalScrollBar")) is not None:
        _scroll.setVerticalScrollBarPolicy(
            (Qt.ScrollBarPolicy.ScrollBarAsNeeded if v else Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
            if isinstance(v, bool) else v
        )
    if (v := kwargs.get("frameShape")) is not None:
        _scroll.setFrameShape(v)
    if (v := kwargs.get("frameShadow")) is not None:
        _scroll.setFrameShadow(v)

    widget(**kwargs, _widget=_scroll)
    return _scroll



from typing import Unpack, NotRequired

from PyQt6.QtCore import Qt
from PyQt6.QtGui import QPixmap
from PyQt6.QtWidgets import QLabel

from .options import Align
from .widget import WidgetKwargs, widget


class LabelKwargs(WidgetKwargs):
    text: NotRequired[str]
    alignment: NotRequired[Align]
    wordWrap: NotRequired[bool]
    indent: NotRequired[int]
    margin: NotRequired[int]
    scaledContents: NotRequired[bool]
    openExternalLinks: NotRequired[bool]
    textFormat: NotRequired[Qt.TextFormat]
    pixmap: NotRequired[QPixmap]


def label(*, _label: QLabel | None = None, **kwargs: Unpack[LabelKwargs]) -> QLabel:
    if _label is None:
        _label = QLabel()

    # ---- QLabel-specific ----
    if (v := kwargs.get("text")) is not None:
        _label.setText(v)

    if (v := kwargs.get("alignment")) is not None:
        _label.setAlignment(v)

    if (v := kwargs.get("wordWrap")) is not None:
        _label.setWordWrap(v)

    if (v := kwargs.get("indent")) is not None:
        _label.setIndent(v)

    if (v := kwargs.get("margin")) is not None:
        _label.setMargin(v)

    if (v := kwargs.get("scaledContents")) is not None:
        _label.setScaledContents(v)

    if (v := kwargs.get("openExternalLinks")) is not None:
        _label.setOpenExternalLinks(v)

    if (v := kwargs.get("textFormat")) is not None:
        _label.setTextFormat(v)

    if (v := kwargs.get("pixmap")) is not None:
        _label.setPixmap(v)

    # ---- apply base QWidget params ----
    widget(**kwargs, _widget=_label)

    return _label


from typing import Callable, Unpack

from PyQt6.QtCore import QSize
from PyQt6.QtGui import QIcon
from PyQt6.QtWidgets import QApplication, QMainWindow, QWidget, QStatusBar, QMenuBar, QToolBar, QLayout

from .widget import WidgetKwargs, widget


class MainWindow(QMainWindow):
    def __init__(
        self,
        *,
        parent: QMainWindow | None = None,
        root: QWidget | QLayout = None,
        statusBar: QStatusBar = None,
        menuBar: QMenuBar = None,
        toolBars: list[QToolBar] = None,
        # Signals
        onWindowTitleChanged: Callable[[str], None] = None,
        onWindowIconChanged: Callable[[QIcon], None] = None,
        onIconSizeChanged: Callable[[QSize], None] = None,
        onStatusBarChanged: Callable[[QStatusBar], None] = None,
        onMenuBarChanged: Callable[[QMenuBar], None] = None,
        onToolBarAdded: Callable[[QToolBar], None] = None,
        onToolBarRemoved: Callable[[QToolBar], None] = None,
        **kwargs: Unpack[WidgetKwargs]
    ):
        super().__init__(parent=parent)

        if root is not None:
            if isinstance(root, QLayout):
                self.setLayout(root)
            if isinstance(root, QWidget):
                self.setCentralWidget(root)

        if statusBar is not None:
            self.setStatusBar(statusBar)

        if menuBar is not None:
            self.setMenuBar(menuBar)

        if toolBars:
            for tb in toolBars:
                self.addToolBar(tb)

        # ---------- Signals ----------
        if onWindowTitleChanged is not None:
            self.windowTitleChanged.connect(onWindowTitleChanged)

        if onWindowIconChanged is not None:
            self.windowIconChanged.connect(onWindowIconChanged)

        if onIconSizeChanged is not None:
            self.iconSizeChanged.connect(onIconSizeChanged)

        if onStatusBarChanged is not None:
            self.statusBarChanged.connect(onStatusBarChanged)

        if onMenuBarChanged is not None:
            self.menuBarChanged.connect(onMenuBarChanged)

        if onToolBarAdded is not None:
            self.toolBarAdded.connect(onToolBarAdded)

        if onToolBarRemoved is not None:
            self.toolBarRemoved.connect(onToolBarRemoved)

        # ---- apply base QWidget params ----
        widget(**kwargs, _widget=self)

    @classmethod
    def execute(cls, argv: list[str] | None = None):
        app = QApplication(argv or [])
        window = cls()
        window.show()
        return app.exec_()



from typing import NotRequired, Unpack, Callable

from PyQt6.QtGui import QIcon
from PyQt6.QtWidgets import QPushButton

from .widget import WidgetKwargs, widget


class ButtonKwargs(WidgetKwargs):
    text: NotRequired[str]
    checkable: NotRequired[bool]
    checked: NotRequired[bool]
    autoDefault: NotRequired[bool]
    default: NotRequired[bool]
    flat: NotRequired[bool]
    icon: NotRequired[str]
    shortcut: NotRequired[str]

    onClicked: NotRequired[Callable[[bool], None]]
    onPressed: NotRequired[Callable[[], None]]
    onReleased: NotRequired[Callable[[], None]]
    onToggled: NotRequired[Callable[[bool], None]]


def button(*, _button: QPushButton | None = None, **kwargs: Unpack[ButtonKwargs]) -> QPushButton:
    if _button is None:
        _button = QPushButton()

    if (v := kwargs.get("text")) is not None:
        _button.setText(v)

    if (v := kwargs.get("checkable")) is not None:
        _button.setCheckable(v)

    if (v := kwargs.get("checked")) is not None:
        _button.setChecked(v)

    if (v := kwargs.get("autoDefault")) is not None:
        _button.setAutoDefault(v)

    if (v := kwargs.get("default")) is not None:
        _button.setDefault(v)

    if (v := kwargs.get("flat")) is not None:
        _button.setFlat(v)

    if (v := kwargs.get("icon")) is not None:
        _button.setIcon(QIcon(v))

    if (v := kwargs.get("shortcut")) is not None:
        _button.setShortcut(v)

    # -------- Handlers (explicit) ---------

    if (fn := kwargs.get("onClicked")) is not None:
        _button.clicked.connect(fn)

    if (fn := kwargs.get("onPressed")) is not None:
        _button.pressed.connect(fn)

    if (fn := kwargs.get("onReleased")) is not None:
        _button.released.connect(fn)

    if (fn := kwargs.get("onToggled")) is not None:
        _button.toggled.connect(fn)

    # ---- apply base QWidget params ----
    widget(**kwargs, _widget=_button)

    return _button
